class FeatAndOffsetGetter(object):
    def get_term_offsets(self, doc):
        return None

    def get_metadata_offsets(self, doc):
        return None

    def get_metadata_row_offsets(self, doc, row):
        return []